import numpy as np


def geocolor(ax, ds, projection):
    """
    Follows
    https://unidata.github.io/python-gallery/examples/mapping_GOES16_TrueColor.html

    Use origin="lower" for imshow because we are using an interpolated grid of data not
    the native layout for data used in the link.
    """
    maxval = 120
    gamma = 2.2

    red = (ds.refl_0_65um_nom / maxval) ** (1 / gamma)
    green = (ds.refl_0_86um_nom / maxval) ** (1 / gamma)
    blue = (ds.refl_0_47um_nom / maxval) ** (1 / gamma)

    true_green = 0.45 * red + 0.1 * green + 0.45 * blue

    color_array = np.dstack([red, true_green, blue])

    x = ds.longitude
    y = ds.latitude

    return ax.imshow(
        color_array,
        origin="lower",
        extent=[x.min(), x.max(), y.min(), y.max()],
        transform=projection,
    )



import re
import pathlib
import datetime

import parse
import numpy as np
import xarray as xr

from . import plot


#: Filename pattern of netCDF files on the AERIS server
nc_filename = (
    "clavrx_OR_ABI-L1b-RadF-M6C01_G16_s"
    "{year:04d}{day:03d}{hour:02d}{minute:02d}"
    "{something}_BARBADOS-2KM-FD.level2.nc"
)

# GOES data parameters
filename_format = (
    "{layer}_{year:04d}-{month:02d}-{day:02d}_{hour:02d}-{minute:02d}.tiff"
)
time_resolution = datetime.timedelta(minutes=10)

# Defaults (for EUREC4A)
default_layer = "GOES-East_ABI_Band2_Red_Visible_1km"
# Spatial resolution (degrees)
default_spatial_resolution = 0.01
# Image extent [S, W, N, E]
default_bbox = [10.0, -60.0, 15.0, -50.0]


def filename_at_time(time, layer=default_layer):
    return filename_format.format(
        layer=layer,
        year=time.year,
        month=time.month,
        day=time.day,
        hour=time.hour,
        minute=time.minute,
    )


def time_from_filename(filename):
    image_info = parse.parse(filename_format, filename).named
    del image_info["layer"]

    return datetime.datetime(**image_info)


def find_images_in_path(path, layer=default_layer):
    # Use a specific layer or use layer "*" to get all layers
    # Replace all times with "*" to search for all images
    filename_to_search = re.sub(r"\{\w+:\d\dd}", "*", filename_format).format(
        layer=layer
    )

    # Use pathlib to search the give path for all matching filenames
    path = pathlib.Path(path)
    return sorted(list(path.rglob(filename_to_search)))


def _load_image(filename):
    from osgeo import gdal

    # Load GeoTiff data
    ds = gdal.Open(filename)

    # Get the dimentions of column and row
    nc = ds.RasterXSize
    nr = ds.RasterYSize

    # Read elevation data
    # The first dimension of the array is RGB but they are all the same so just return
    # the first on for now. May need to change this if we can get colour satellite
    # images
    data = ds.ReadAsArray()[0]

    # Get Longitude and Latitude info
    geotransform = ds.GetGeoTransform()
    x_origin = geotransform[0]
    y_origin = geotransform[3]
    pixel_width = geotransform[1]
    pixel_height = geotransform[5]

    # Generate Longitude and Latitude array
    lons = x_origin + np.arange(0, nc) * pixel_width
    lats = y_origin + np.arange(0, nr) * pixel_height

    return lons, lats, data


def load_nc(path, time):
    """Load the netCDF dataset corresponding to the given time

    This function finds files matching the AERIS :data:`nc_filename` formatted for the
    given time

    Args:
        path (str): The directory containing GOES netCDF files
        time (datetime.datetime): The time of the file to load

    Returns:
        xarray.dataset: The loaded netCDF file with values filtered where the
            coordinates are NaN

    Raises:
        FileNotFoundError: If the netCDF is not present

        FileExistsError: If multiple matching netCDF files are found with the same name
    """
    filename = nc_filename.format(
        year=time.year,
        day=time.timetuple().tm_yday,
        hour=time.hour,
        minute=time.minute,
        something="*",
    )

    # Find matching filenames in the GOES folder
    file_path = list(pathlib.Path(path).rglob(filename))

    if len(file_path) == 0:
        raise FileNotFoundError("No GOES data found in {} for {}".format(path, time))
    elif len(file_path) > 1:
        raise FileExistsError(
            "More than one file found in {} for {}".format(path, time)
        )

    dataset = xr.load_dataset(str(file_path[0]))

    # Remove values where the coordinates are NaNs
    dataset = dataset.where(~dataset.longitude.isnull(), drop=True)

    return dataset



import matplotlib.pyplot as plt
from metpy.plots import SkewT
import dateutil.parser

from .. import load_flight


def main(
    flight_data_path, filter_by=dict(time_interval=(None, None)), plot_type="skewt"
):
    ds = load_flight(flight_data_path, debug=True)

    if "time_interval" in filter_by:
        ds_filtered = ds.sel(Time=slice(*filter_by["time_interval"]))
        if ds_filtered.Time.count() == 0:
            raise Exception(
                "Nothing found in selected time range, data"
                " time range: [{} : {}]".format(
                    ds.Time.min().values, ds.Time.max().values
                )
            )
        else:
            ds = ds_filtered
        title = "flight {}, extract: {} to {}".format(
            ds.flight_number, *filter_by["time_interval"]
        )
    elif "leg" in filter_by:
        raise NotImplementedError("Filter by leg not currently implemented")
        # ds = _filter_by_flight_leg(ds, flight_data_path, *filter_by['leg'])
        title = "leg: {} {}".format(*filter_by["leg"])
    else:
        raise NotImplementedError

    alt, p, T, Td, u, v = _extract_plot_data(ds)

    if plot_type == "skewt":
        skewt(p, T, Td, u, v)
    elif plot_type == "linear_altitude":
        _plot_linear_altitude(ds)
    else:
        raise NotImplementedError(plot_type)

    plt.title(title)
    plt.show()


def _plot_linear_altitude(ds):
    fig, ax = plt.subplots()

    ds_ = ds.swap_dims(dict(Time="ALT_OXTS"))

    ds_.TAT_ND_R.plot(y="ALT_OXTS", ax=ax, color="green")
    ds_.TDEW_BUCK.plot(y="ALT_OXTS", ax=ax, color="red")


def _extract_plot_data(ds):
    p = ds.PS_AIR
    T = ds.TAT_ND_R - 273.15
    Td = ds.TDEW_BUCK - 273.15
    u = ds.U_OXTS
    v = ds.V_OXTS
    alt = ds.HGT_RADR1

    return alt, p, T, Td, u, v


def skewt(p, T, Td, u, v):
    """

    Adapted from the Metpy advanced sounding example
    (https://unidata.github.io/MetPy/latest/examples/Advanced_Sounding.html#sphx-glr-examples-advanced-sounding-py)

    """
    fig = plt.figure(figsize=(9, 9))
    skew = SkewT(fig)

    # Plot the data using normal plotting functions, in this case using
    # log scaling in Y, as dictated by the typical meteorological plot.
    skew.plot(p, T, "r")
    skew.plot(p, Td, "g")

    # Calculate LCL height and plot as black dot. Because `p`'s first value is
    # ~1000 mb and its last value is ~250 mb, the `0` index is selected for
    # `p`, `T`, and `Td` to lift the parcel from the surface. If `p` was inverted,
    # i.e. start from low value, 250 mb, to a high value, 1000 mb, the `-1` index
    # should be selected.
    # lcl_pressure, lcl_temperature = mpcalc.lcl(p[0], T[0], Td[0])
    # skew.plot(lcl_pressure, lcl_temperature, 'ko', markerfacecolor='black')

    # Calculate full parcel profile and add to plot as black line
    # prof = mpcalc.parcel_profile(p, T[0], Td[0]).to('degC')
    # skew.plot(p, prof, 'k', linewidth=2)

    # Shade areas of CAPE and CIN
    # skew.shade_cin(p, T, prof, Td)
    # skew.shade_cape(p, T, prof)

    # An example of a slanted line at constant T -- in this case the 0
    # isotherm
    # skew.ax.axvline(0, color='c', linestyle='--', linewidth=2)

    # Add the relevant special lines
    skew.plot_dry_adiabats()
    skew.plot_moist_adiabats()
    skew.plot_mixing_lines()

    skew.ax.set_ylim(1050, p.min())
    skew.ax.set_xlim(-10, 30)

    return fig


if __name__ == "__main__":

    def _flight_leg(s):
        leg_type, leg_number = s.split(":")
        return leg_type, int(leg_number)

    import argparse

    argparser = argparse.ArgumentParser()
    argparser.add_argument("flight_data_path")
    argparser.add_argument(
        "--time-interval",
        help="e.g. 2020-01-17T14:00Z 2020-01-17T15:00Z",
        nargs=2,
        type=dateutil.parser.parse,
    )
    argparser.add_argument("--leg", help="e.g. profile:0", type=_flight_leg)
    argparser.add_argument(
        "--type", choices=["skewt", "linear_altitude"], default="skewt"
    )

    args = argparser.parse_args()

    kws = dict(plot_type=args.type)

    if args.time_interval is not None:
        kws["filter_by"] = dict(time_interval=args.time_interval)
    elif args.leg:
        kws["filter_by"] = dict(leg=args.leg)
    else:
        raise NotImplementedError

    main(args.flight_data_path, **kws)



import datetime


def round_datetime(time, resolution, mode=None):
    """Round the time by resolution

    Args:
        time (datetime.datetime | datetime.timedelta):
        resolution (datetime.timedelta):
        mode (): Round up or down (ceil or floor)

    Returns:
        datetime.datetime
    """
    try:
        excess = time % resolution
    except TypeError:
        excess = (
            datetime.timedelta(
                days=time.day,
                hours=time.hour,
                minutes=time.minute,
                seconds=time.second,
                microseconds=time.microsecond,
            )
            % resolution
        )

    if mode == "ceil" or (excess > resolution / 2 and mode != "floor"):
        # Round up
        return time - excess + resolution
    else:
        # Round down
        return time - excess



"""
::

    Usage:
        flight_track_frames.py  <flight_data_path>
            [<lon_min> <lon_max> <lat_min> <lat_max> <resolution>]
            [--goes_path=<path>]
            [--output_path=<path>]
        flight_track_frames.py  (-h | --help)

    Arguments:
        <flight_data_path>  Input flight data

    Options:
        -h --help           Show help
        --goes_path=<path>
            Folder containing downloaded GOES images [default: .]
        --output_path=<path>
            Folder to put the output frames in [default: .]

"""

import datetime

import numpy as np
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
from scipy.interpolate import griddata
import xarray as xr

from .. import load_flight, plots, util
from ..util import scripting
from ..external import eurec4a, goes


def main():
    scripting.parse_docopt_arguments(generate, __doc__)
    return


def generate(
    flight_data_path,
    lon_min=-60,
    lon_max=-56.4,
    lat_min=12,
    lat_max=14.4,
    resolution=0.01,
    goes_path=".",
    output_path=".",
):
    substep = datetime.timedelta(minutes=1)

    # Setup the grid to interpolate the satellite data on to
    lon = np.arange(lon_min, lon_max, resolution)
    lat = np.arange(lat_min, lat_max, resolution)
    lon_grid, lat_grid = np.meshgrid(lon, lat)

    # Load flight data
    dataset = load_flight(flight_data_path)

    # Get start and end time for satellite data from flight
    start = dataset.Time[0].data.astype("M8[ms]").astype("O")[()]
    end = dataset.Time[-1].data.astype("M8[ms]").astype("O")[()]

    # Start the satellite images at the nearest time to the flight start
    sat_image_time = util.round_datetime(start, goes.time_resolution)

    # Loop over satellite images
    n = 0
    # Start on the minute
    time = util.round_datetime(start, datetime.timedelta(minutes=1), mode="ceil")
    while time <= end:
        # Load the current satellite image
        goes_data = goes.load_nc(goes_path, sat_image_time)

        # Interpolate the satellite data to a regular grid
        goes_data_grid = xr.Dataset(coords=dict(latitude=lat, longitude=lon))
        for band in ["refl_0_65um_nom", "refl_0_86um_nom", "refl_0_47um_nom"]:
            band_grid = griddata(
                (
                    goes_data["longitude"].values.flatten(),
                    goes_data["latitude"].values.flatten(),
                ),
                goes_data[band].values.flatten(),
                (lon_grid, lat_grid),
            )

            goes_data_grid[band] = (["latitude", "longitude"], band_grid)

        sat_image_time += goes.time_resolution
        while time < sat_image_time - goes.time_resolution / 2 and time <= end:
            fig, ax = make_frame(goes_data_grid)

            overlay_flight_path_segment(ax, dataset, time)

            path_fig = (
                output_path
                + "/"
                + "flight{}_track_frame_{:03d}.png".format(
                    dataset.attrs["flight_number"], n
                )
            )
            plt.text(
                0, 0, str(time), transform=ax.transAxes, fontdict=dict(color="green")
            )
            plt.savefig(path_fig, bbox_inches="tight")
            print("Saved flight track to `{}`".format(str(path_fig)))
            plt.close()

            time += substep
            n += 1


def make_frame(goes_data):
    # create figure
    bbox = [-60, -56.4, 12, 14.4]
    domain_aspect = (bbox[3] - bbox[2]) / (bbox[1] - bbox[0])
    fig = plt.figure(figsize=(11.0, domain_aspect * 10), dpi=96)
    projection = ccrs.PlateCarree()
    ax = plt.axes(projection=projection)

    plots.add_land_and_sea(ax)

    # Plot the current satellite image
    goes.plot.geocolor(ax, goes_data, projection)

    eurec4a.add_halo_circle(ax, color="teal", linewidth=3)

    ax.set_extent(bbox, crs=projection)

    return fig, ax


def overlay_flight_path_segment(ax, flight_data, time):
    # Plot the flight track +- the satellite resolution
    start = time - goes.time_resolution
    end = time + goes.time_resolution

    # Plot the full flight path in a faded red
    plots.flight_path(
        ax=ax,
        ds=flight_data,
        vmin=-10,
        vmax=0,
        cmap="Reds",
        alpha=0.3,
        linewidths=3,
        add_cmap=False,
    )

    # Plot the +-10 mins of flight path normally
    plots.flight_path(
        ax=ax,
        ds=flight_data.sel(Time=slice(start, end)),
        cmap="cool",
        mark_end_points=False,
    )

    # Add a marker with the current position and rotation
    plots.add_flight_position(ax, flight_data.sel(Time=time))


if __name__ == "__main__":
    main()



"""
> python -m twinotter.plots.basic_flight_track /path/to/data
"""

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
from pathlib import Path
from datetime import datetime

from .. import load_flight
from .. import plots
from ..external import eurec4a


def main():
    import argparse

    argparser = argparse.ArgumentParser()
    argparser.add_argument("flight_data_path", nargs="+")
    argparser.add_argument("--show-gui", default=False, action="store_true")

    args = argparser.parse_args()

    for flight_data_path in args.flight_data_path:
        generate(flight_data_path=flight_data_path, show_gui=args.show_gui)


def generate(flight_data_path, show_gui=False):
    flight_data_path = Path(flight_data_path)

    # create figure
    bbox = [-60, -56.4, 12, 14.4]
    domain_aspect = (bbox[3] - bbox[2]) / (bbox[1] - bbox[0])
    fig = plt.figure(figsize=(11.0, domain_aspect * 10), dpi=96)
    ax = plt.axes(projection=ccrs.PlateCarree())
    add_features(ax)
    ax.set_extent(bbox, crs=ccrs.PlateCarree())
    fig.tight_layout()

    ds = load_flight(flight_data_path, debug=True)
    plots.flight_path(ax=ax, ds=ds)

    fig = ax.figure
    caption = "created {} from {}".format(datetime.now(), ds.source_file)
    ax.text(0.0, 0.0, caption, transform=fig.transFigure)

    if flight_data_path.is_file():
        path_fig = flight_data_path.parent
    else:
        path_fig = flight_data_path
    path_fig = (
        path_fig / "figures" / "flight{}_track_altitude.png".format(ds.flight_number)
    )
    path_fig.parent.mkdir(exist_ok=True, parents=True)

    if show_gui:
        plt.show()
    else:
        plt.savefig(str(path_fig), bbox_inches="tight")
        print("Saved flight track to `{}`".format(str(path_fig)))


def add_features(ax):
    # Shade land and sea
    plots.add_land_and_sea(ax)

    # Add HALO circle
    eurec4a.add_halo_circle(ax)

    return


if __name__ == "__main__":
    main()



from pathlib import Path

from tqdm import tqdm
import matplotlib.pyplot as plt

from .. import load_flight, load_segments


colors = {
    "level": "cyan",
    "profile": "magenta",
}


def main():
    import argparse

    argparser = argparse.ArgumentParser()
    argparser.add_argument("flight_data_path")
    argparser.add_argument("flight_segments_file")
    argparser.add_argument("--show-gui", default=False, action="store_true")
    argparser.add_argument("--output_path", default=None)
    args = argparser.parse_args()

    generate(
        args.flight_data_path,
        args.flight_segments_file,
        show_gui=args.show_gui,
        output_path=args.output_path,
    )


def generate(flight_data_path, flight_segments_file, show_gui=False, output_path=None):
    ds = load_flight(flight_data_path)
    segments = load_segments(flight_segments_file)

    # Produce the basic time-height plot
    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()
    ax1.plot(ds.Time, ds.ALT_OXTS / 1000, color="k", alpha=0.5)
    ax1.set_ylabel("Altitude (km)")

    ax2.plot(ds.Time, ds.ROLL_OXTS, color="k", linestyle="--", alpha=0.1)
    ax2.set_ylabel("Roll Angle")

    # For each segment overlay a coloured line onto the time-height plot
    for segment in tqdm(segments["segments"]):
        ds_section = ds.sel(Time=slice(segment["start"], segment["end"]))
        label = segment["kinds"][0]

        linestyle = "-"
        try:
            color = colors[label]
        except KeyError:
            # If the primary segment type doesn't have an assigned colour but one of the
            # other "kinds" matches then use that colour with a dashed line
            color = "yellow"
            linestyle = "--"
            for segment_type in colors:
                if segment_type in segment["kinds"]:
                    color = colors[segment_type]

        ax1.plot(
            ds_section.Time,
            ds_section.ALT_OXTS / 1000,
            linestyle=linestyle,
            color=color,
            linewidth=2,
            alpha=0.75,
        )

    if hasattr(ax1, "secondary_yaxis"):
        # `ax.secondary_yaxis` was added in matplotlib v3.1
        ax1_fl = ax1.secondary_yaxis(
            location=-0.15, functions=(lambda y: (y * 1000 * 3.281) / 100, lambda x: x)
        )
        ax1_fl.set_ylabel(r"Flight level [100ft]")

    for label in ax1.get_xmajorticklabels():
        label.set_rotation(30)
        label.set_horizontalalignment("right")

    if show_gui:
        plt.show()
    else:
        if output_path is None:
            output_path = (
                Path(flight_data_path) / "figures" / "height-time-with-legs.png"
            )
        else:
            output_path = Path(output_path)

        output_path.parent.mkdir(exist_ok=True)
        plt.savefig(str(output_path), bbox_inches="tight")


if __name__ == "__main__":
    main()



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from mpl_toolkits.mplot3d import Axes3D
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import xarray as xr


def flight_path(
    ax,
    ds,
    vmin=0,
    vmax=3,
    cmap_steps=12,
    cmap="jet",
    transform=ccrs.PlateCarree(),
    add_cmap=True,
    mark_end_points=True,
    **kwargs
):

    lc = colored_line_plot(
        ax,
        ds.LON_OXTS,
        ds.LAT_OXTS,
        ds.ALT_OXTS / 1000,
        vmin=vmin,
        vmax=vmax,
        cmap_steps=cmap_steps,
        cmap=cmap,
        transform=transform,
        **kwargs
    )

    if add_cmap:
        cbar = plt.colorbar(lc, ax=ax)
        cbar.set_label("Altitude (km)")

    ax.set_xlabel(xr.plot.utils.label_from_attrs(ds.LON_OXTS))
    ax.set_ylabel(xr.plot.utils.label_from_attrs(ds.LAT_OXTS))

    # Add marker for start and end positions
    if mark_end_points:
        ax.text(ds.LON_OXTS[0], ds.LAT_OXTS[0], "S", transform=ccrs.PlateCarree())
        ax.text(ds.LON_OXTS[-1], ds.LAT_OXTS[-1], "F", transform=ccrs.PlateCarree())

    return


def flight_path_3d(ds, ax=None):
    if ax == None:
        ax = plt.gca(projection="3d")

    ax.plot(ds.LON_OXTS, ds.LAT_OXTS, zs=0, zdir="z", color="grey")

    ax.plot(
        ds.LON_OXTS, ds.ALT_OXTS, zs=ds.LAT_OXTS.max() + 0.1, zdir="y", color="grey"
    )

    ax.plot(
        ds.LAT_OXTS, ds.ALT_OXTS, zs=ds.LON_OXTS.max() + 0.1, zdir="x", color="grey"
    )

    ax.plot(ds.LON_OXTS, ds.LAT_OXTS, ds.ALT_OXTS, color="k", lw=3, zorder=10)


def colored_line_plot(
    ax, x, y, color, vmin=None, vmax=None, cmap="gray", cmap_steps=0, **kwargs
):
    """Add a multicolored line to an existing plot

    Args:
        x (np.array): The x points of the plot

        y (np.array): The y points of the plot

        color (np.array): The color of the line at the xy points

        vmin (scalar, optional): The minimum of the colorscale. Defaults to the
            minimum of the color array.

        vmax (scalar, optional): The maximum of the colorscale. Defaults to the
            maximum of the color array.

        cmap (str, optional): Colormap to plot. Default is grey.

        cmap_steps (int, optional): Number of discrete steps in the colorscale.
            Defaults is zero for a continuous colorscale.

        kwargs: Other keyword arguments to pass to LineCollection
    returns:
        matplotlib.collections.LineCollection:
            The plotted LineCollection. Required as argument to
            :py:func:`matplotlib.pyplot.colorbar`
    """
    # Set the color scalings
    if vmin is None:
        vmin = color.min()
    if vmax is None:
        vmax = color.max()

    # Break the xy points up in to line segments
    segments = np.array([(x[:-1].values, x[1:].values), (y[:-1].values, y[1:].values)])
    segments = np.transpose(segments, axes=(2, 1, 0))

    # Create discretised colourmap
    cmap = plt.get_cmap(cmap)
    if cmap_steps != 0:
        cmap = mpl.colors.ListedColormap(
            [cmap(n / (cmap_steps - 1)) for n in range(cmap_steps)]
        )

    # Collect the line segments
    lc = LineCollection(segments, cmap=cmap, norm=plt.Normalize(vmin, vmax), **kwargs)

    # Set the line color to the specified array
    lc.set_array(color)

    # Add the colored line to the existing plot
    ax.add_collection(lc)

    # autoscale if limits haven't already been set so that the linecollection
    # is visible
    if ax.get_xlim() == (0, 1) and ax.get_ylim() == (0, 1):
        ax.autoscale()

    return lc


def add_land_and_sea(ax):
    # Shade land and sea
    ax.imshow(
        np.tile(
            np.array([[cfeature.COLORS["water"] * 255]], dtype=np.uint8), [2, 2, 1]
        ),
        origin="upper",
        transform=ccrs.PlateCarree(),
        extent=[-180, 180, -180, 180],
    )

    ax.add_feature(
        cfeature.NaturalEarthFeature(
            "physical",
            "land",
            "10m",
            edgecolor="black",
            facecolor=cfeature.COLORS["land"],
        )
    )

    ax.gridlines(linestyle="--", color="black", draw_labels=True)

    return


def add_flight_position(ax, dataset):
    ax.plot(
        dataset.LON_OXTS,
        dataset.LAT_OXTS,
        marker=(2, 0, -float(dataset.HDG_OXTS)),
        color="red",
    )
    ax.plot(
        dataset.LON_OXTS,
        dataset.LAT_OXTS,
        marker=(3, 0, -float(dataset.HDG_OXTS)),
        color="red",
    )
    return



"""Functionality related to the EUREC4A field campaign from 2020
"""

import requests
import yaml
import cartopy.crs as ccrs
import matplotlib.patches as mpatches

#: A dictionary of standard colours for the various platforms
colors = dict(
    ATR="DeepOrange",
    Atalante="Fuchsia",
    BCO="royalblue",
    HALO="teal",
    P3="grey",
    TO="tab:red",
    MSM="indigo",
    Meteor="darkblue",
    RHB="deepskyblue",
)

# HALO circle attributes
lat = 13 + (18 / 60)
lon = -(57 + (43 / 60))
r = 1


def _compute_radius(ortho, radius_degrees):
    phi1 = lat + radius_degrees if lat <= 0 else lat - radius_degrees
    _, y1 = ortho.transform_point(lon, phi1, ccrs.PlateCarree())
    return abs(y1)


def add_halo_circle(ax, color=colors["HALO"], alpha=0.3, **kwargs):
    """Add the HALO circle to the current

    Projection of the HALO circle adapted from AJDawson's answer on stackoverflow
    https://stackoverflow.com/questions/52105543/drawing-circles-with-cartopy-in-orthographic-projection

    Args:
        ax ():
        color (str): The circle of the colour. Any matplotlib compatible colour. Default
            is teal (set in the dictionary :data:colors:).
        alpha (float): The transparancy of the circle (between 0 and 1). Default is 0.3
        **kwargs: Other keywords to pass to :meth:matplotlib.axes.Axes.add_patch:
    """
    # Define the projection used to display the circle:
    proj = ccrs.Orthographic(central_longitude=lon, central_latitude=lat)

    # Compute the required radius in projection native coordinates:
    r_ortho = _compute_radius(proj, r)
    ax.add_patch(
        mpatches.Circle(
            xy=[lon, lat],
            radius=r_ortho,
            transform=proj,
            zorder=30,
            fill=False,
            color=color,
            alpha=alpha,
            **kwargs
        )
    )


TO_segs_url = (
    "https://raw.githubusercontent.com/"
    "EUREC4A-UK/flight-phase-separation/twinotter/"
    "flight_phase_files/TO/EUREC4A_TO_Flight-Segments_{}_0.1.yaml"
)
segs_url = (
    "https://raw.githubusercontent.com/"
    "eurec4a/flight-phase-separation/master/"
    "flight_phase_files/"
)

HALO_flight_numbers = [
    119,
    122,
    124,
    126,
    128,
    130,
    131,
    202,
    205,
    207,
    209,
    211,
    213,
    215,
    218,
]

P3_flight_numbers = [117, 119, 123, 124, 131, 203, 204, 205, 209, 210, 211]
flight_segs_urls = dict(
    TO={
        330: TO_segs_url.format("20200124a"),
        331: TO_segs_url.format("20200124b"),
        332: TO_segs_url.format("20200126a"),
        333: TO_segs_url.format("20200126b"),
        334: TO_segs_url.format("20200128a"),
        335: TO_segs_url.format("20200128b"),
        336: TO_segs_url.format("20200130a"),
        337: TO_segs_url.format("20200131a"),
        338: TO_segs_url.format("20200131b"),
        339: TO_segs_url.format("20200202a"),
        340: TO_segs_url.format("20200205a"),
        341: TO_segs_url.format("20200205b"),
        342: TO_segs_url.format("20200206a"),
        343: TO_segs_url.format("20200207a"),
        344: TO_segs_url.format("20200207b"),
        345: TO_segs_url.format("20200209a"),
        346: TO_segs_url.format("20200209b"),
        347: TO_segs_url.format("20200210a"),
        348: TO_segs_url.format("20200211a"),
        349: TO_segs_url.format("20200211b"),
        350: TO_segs_url.format("20200213a"),
        351: TO_segs_url.format("20200213b"),
        352: TO_segs_url.format("20200214a"),
        353: TO_segs_url.format("20200215a"),
        354: TO_segs_url.format("20200215b"),
    },
    HALO={
        flight_number: segs_url
        + "HALO/EUREC4A_HALO_Flight-Segments_20200{}.yaml".format(flight_number)
        for flight_number in HALO_flight_numbers
    },
    P3={
        flight_number: segs_url
        + "P3/EUREC4A_ATOMIC_P3_Flight-segments_20200{}_v0.5.yaml".format(flight_number)
        for flight_number in P3_flight_numbers
    },
)


def load_segments(flight_number, platform="TO"):
    """Load flight segments yaml file from EUREC4A github repository

    See https://github.com/eurec4a/flight-phase-separation

    Args:
        flight_number (int):
        platform (str): The short name for the platforms used in EUREC4A. Currently
            either "TO", "HALO", or "P3".

    Returns:
        dict:

    """
    yaml_file = requests.get(flight_segs_urls[platform][flight_number]).text

    return yaml.safe_load(yaml_file)



"""Quicklook plots for each segment over a single flight.

Use the flight-segments .yaml produced from
:mod:`twinotter.plots.interactive_flight_track`


Usage::

    $ python -m twinotter.quicklook <flight_data_path> <flight_segments_file>

"""

from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt
import scipy.constants
import metpy.calc
from metpy.units import units

from . import load_flight, load_segments, count_segments, extract_segments, derive
from .plots import vertical_profile


def main():
    import argparse

    argparser = argparse.ArgumentParser()
    argparser.add_argument("flight_data_path")
    argparser.add_argument("flight_segments_file")

    args = argparser.parse_args()

    generate(
        flight_data_path=args.flight_data_path,
        flight_segments_file=args.flight_segments_file,
    )

    return


def generate(flight_data_path, flight_segments_file):
    ds = load_flight(flight_data_path)
    flight_segments = load_segments(flight_segments_file)

    # Quicklook plots for the full flight
    figures = plot_level(ds)
    savefigs(figures, ds.attrs["flight_number"], "", "")

    plot_individual_phases(ds, flight_segments, "level", plot_level)
    plot_individual_phases(ds, flight_segments, "profile", plot_profile)

    # Make a combined plot of all profiles
    profiles = extract_segments(ds, flight_segments, "profile")
    figures = plot_profile(profiles)
    savefigs(figures, ds.attrs["flight_number"], "profile", "_combined")


def plot_individual_phases(ds, flight_segments, segment_type, plot_func):
    for n in range(count_segments(flight_segments, segment_type)):
        ds_section = extract_segments(ds, flight_segments, segment_type, n)
        figures = plot_func(ds_section)
        savefigs(figures, ds.attrs["flight_number"], segment_type, n)


def plot_level(ds):
    figures = []

    fig, axes = plt.subplots(nrows=5, ncols=1, sharex="all", figsize=[16, 15])

    # Temperature and Dewpoint
    axes[0].plot(ds.Time, ds.TAT_ND_R, label=r"True")
    axes[0].plot(ds.Time, ds.TDEW_BUCK, label=r"Dewpoint")
    axes[0].set_ylabel("Temperature (K)")
    axes[0].legend()

    # Velocities
    axes[1].plot(ds.Time, ds.U_OXTS, label=r"Zonal")
    axes[1].plot(ds.Time, ds.V_OXTS, label=r"Meridional")
    axes[1].set_ylabel("Velocity (m s$^{-1}$)")
    axes[1].legend()

    axes[2].plot(ds.Time, ds.W_OXTS, label=r"Vertical")
    axes[2].set_ylabel("Vertical Velocity (m s$^{-1}$)")

    # Shortwave radiation
    axes[3].plot(ds.Time, ds.SW_DN_C, label=r"SW Downwelling")
    axes[3].plot(ds.Time, ds.SW_UP_C, label=r"SW Upwelling")

    # Longwave radiation
    axes[3].plot(ds.Time, ds.LW_DN_C, label=r"LW Downwelling")
    axes[3].plot(ds.Time, ds.LW_UP_C, label=r"LW Upwelling")

    axes[3].set_ylabel("Irradiance (W m$^{-2}$)")
    axes[3].legend()

    # CPC Concentration
    axes[4].plot(ds.Time, ds.CPC_CONC)
    axes[4].set_ylabel("CPC Concentration (m$^{-3}$)")

    figures.append([fig, "quicklook"])

    # Not all flights have LICOR data
    try:
        fig, axes = plt.subplots(nrows=2, ncols=1, sharex="all", figsize=[16, 15])
        # LICOR data
        axes[0].plot(ds.Time, ds.CO2_LICOR, label=r"CO$_2$ LICOR")
        axes[0].plot(ds.Time, ds.H2O_LICOR, label=r"H$_2$O LICOR")
        axes[0].set_ylabel("Mole Fraction")
        axes[0].legend()

        axes[1].plot(ds.Time, derive.specific_humidity(ds))
        axes[1].set_ylabel("Specific Humidity")

        figures.append([fig, "quicklook_LICOR"])
    except AttributeError:
        pass

    fig, ax = plt.subplots(figsize=[8, 5])

    ax.scatter(
        derive.calculate("equivalent_potential_temperature", ds),
        metpy.calc.specific_humidity_from_dewpoint(ds.PS_AIR, ds.TDEW_BUCK),
        alpha=0.1,
    )
    ax.set_xlabel(r"$\theta_e$")
    ax.set_ylabel("Specific Humidity")
    figures.append([fig, "paluch"])

    return figures


def plot_profile(dataset):
    p = dataset.PS_AIR
    T = dataset.TAT_ND_R
    Td = dataset.TDEW_BUCK
    u = dataset.U_OXTS
    v = dataset.V_OXTS
    fig1 = vertical_profile.skewt(
        p, T - scipy.constants.zero_Celsius, Td - scipy.constants.zero_Celsius, u, v
    )

    theta = metpy.calc.potential_temperature(p * units("hPa"), T * units("K"))
    fig2 = plt.figure()
    plt.plot(theta, dataset.ALT_OXTS, "k.")
    plt.xlim(295, 320)
    plt.ylim(0, 4000)
    plt.xlabel("Potential Temperature (K)")
    plt.ylabel("Altitude (m)")
    plt.title("Flight {}".format(dataset.attrs["flight_number"]))

    rh = metpy.calc.relative_humidity_from_dewpoint(T * units("K"), Td * units("K"))
    fig3 = plt.figure()
    plt.plot(rh, dataset.ALT_OXTS, "b.")
    plt.xlim(0, 1)
    plt.ylim(0, 4000)
    plt.xlabel("Relative Humidity")
    plt.ylabel("Altitude (m)")
    plt.title("Flight {}".format(dataset.attrs["flight_number"]))

    return [(fig1, "skewt"), (fig2, "theta_0-4km"), (fig3, "rh_0-4km")]


def savefigs(figures, flight_number, label, n):
    for fig, figname in figures:
        fn = "flight{}_{}{}_{}.png".format(flight_number, label, n, figname)
        print(fn)
        fig.savefig(fn)
        plt.close(fig)


if __name__ == "__main__":
    main()



"""
Module to work with CSV files of measurements of Cloud Condensation Nuclei
measurements from the CCN-100 probe

"A DMT single-column cloud condensation counter model measures the spectrum of
CCN concentration as a function of supersaturation continuously using
uninterrupted flow and a multichannel, optical particle counter that measures
the size of the activated droplet."
"""

import xarray as xr
import pandas as pd
from pathlib import Path


def _load_meta(fn, nlines=3):
    meta = {}
    with open(fn) as fh:
        for n in range(3):
            line = fh.readline()
            key, value = line.split(",")
            meta[key.strip().lower()] = value.strip()
    return meta


def load_csv(fn, n_header_lines=3):
    """
    Load CCN datafile with filename `fn` and return as xarray.Dataset
    """
    meta = _load_meta(fn, nlines=n_header_lines)
    df = pd.read_csv(fn, skiprows=n_header_lines)
    # cleanup column names
    df.columns = [s.strip().lower().replace(" ", "_") for s in df.columns]
    # make times into datetimes
    df["time"] = meta["date"] + "T" + df["time"] + "Z"
    df["time"] = pd.to_datetime(df["time"], format="%m/%d/%yT%H:%M:%SZ", utc=True)

    ds = xr.Dataset.from_dataframe(df)
    ds = ds.swap_dims(dict(index="time"))

    ds.attrs.update(meta)

    return ds


def load_all(data_path):
    """
    Load all CCN files in `data_path` and combine into single xarray.Dataset
    """
    all_ds = [load_csv(fn=fn) for fn in Path(data_path).glob("*.csv")]
    if len(all_ds) == 0:
        raise Exception("No CCN data found in `{}`".format(data_path))
    else:
        return xr.concat(all_ds, dim="time")



"""Interactive flight track plot

Plot a variable of interest and the flight track then click on either figure
to mark the corresponding points on both figures.

.. code-block:: console

    $ python -m twinotter.plots.interactive_flight_track /path/to/data
"""

import datetime
import tkinter
from tkinter import filedialog, ttk

import yaml
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import SpanSelector
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import pandas as pd

from .. import load_flight
from . import flight_path


yaml.Dumper.ignore_aliases = lambda *args: True

masin_date_format = "{year:04d}{month:02d}{day:02d}"
masin_time_format = "{hour:02d}:{minute:02d}:{second:02d} UTC"

yaml_file_format = (
    "EUREC4A_TO_Flight-Segments_{year:04d}{month:02d}{day:02d}_{version}.yaml"
)


def main():
    import argparse

    argparser = argparse.ArgumentParser()
    argparser.add_argument("flight_data_path")

    args = argparser.parse_args()

    ds = load_flight(flight_data_path=args.flight_data_path)

    root = tkinter.Tk()
    app = FlightPhaseGenerator(ds, root)
    app.mainloop()
    root.destroy()

    return


class FlightPhaseGenerator(tkinter.Frame):
    def __init__(self, ds, parent, *args, **kwargs):
        tkinter.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        self.grid()

        self.ds = ds
        self.flight_information = flight_information(ds)

        # Use datetime functionality as xarray and pandas makes this difficult
        self.time = pd.to_datetime(self.ds.Time.values).to_pydatetime()

        # Flight leg times will be recorded as time since the start of the flight day
        self.flight_day_start = pd.to_datetime(
            ds.Time[0].dt.floor("D").data
        ).to_pydatetime()

        entries = {}
        for n, entry_label in enumerate(self.flight_information):
            label = ttk.Label(self, text=entry_label)
            entry = ttk.Entry(self)

            entry.insert(tkinter.END, self.flight_information[entry_label])

            label.grid(row=n, column=0)
            entry.grid(row=n, column=1)

            entries[entry_label] = entry

        self.quit_button = ttk.Button(self, text="Start", command=self.start)
        self.quit_button.grid(row=n + 1, column=0, columnspan=2)

    def start(self):
        self.root = tkinter.Tk()
        self.root.wm_title(
            "Interactive Flight Track: Flight {}".format(self.ds.attrs["flight_number"])
        )

        # Plot the main variable of interest
        # Change this to whatever variable you want or add additional figures here
        fig, self.ax1 = plt.subplots()
        self.ax1.plot(self.ds.Time, self.ds.ROLL_OXTS, linestyle="--", alpha=0.5)
        self.ax1.set_label("Roll Angle")
        self.ax2 = self.ax1.twinx()
        self.ax2.plot(self.ds.Time, self.ds.ALT_OXTS / 1000)
        self.ax2.set_ylabel("Altitude (km)")

        fig.tight_layout()

        # Add the figures to the TK window
        self.figure_area = tkinter.Frame(self.root)
        self.figure_area.grid(row=0, column=0, columnspan=2)

        canvas = FigureCanvasTkAgg(fig, master=self.figure_area)
        canvas.draw()
        canvas.get_tk_widget().grid(row=0, column=0, columnspan=2)

        # Add an area for buttons beneath the figures
        self.button_area = tkinter.Canvas(self.root)
        self.button_area.grid(row=1, column=1)

        save_button = tkinter.Button(
            master=self.button_area, text="Save", command=self.save
        )
        save_button.grid(row=0, column=0)

        self.quit_button = tkinter.Button(
            master=self.button_area, text="Quit", command=self.quit
        )
        self.quit_button.grid(row=0, column=1)

        # Use an Entry textbox to label the legs
        self.textbox = tkinter.Entry(master=self.root)
        self.textbox.grid(row=1, column=0)

        self.selector = SpanSelector(
            self.ax2, self.highlight_leg, direction="horizontal"
        )

    def save(self):
        year = self.flight_day_start.year
        month = self.flight_day_start.month
        day = self.flight_day_start.day
        filename = filedialog.asksaveasfilename(
            initialfile=yaml_file_format.format(
                year=year, month=month, day=day, version="0.1"
            )
        )
        with open(filename, "w") as f:
            f.write(
                yaml.dump(
                    self.flight_information, default_flow_style=False, sort_keys=False
                )
            )

    # Add a span selector to the time-height plot to highlight legs
    # Drag mouse from the start to the end of a leg and save the corresponding
    # times
    def highlight_leg(self, start, end):
        self.ax2.axvspan(start, end, alpha=0.25, color="r")
        start = _convert_wacky_date_format(start)
        end = _convert_wacky_date_format(end)

        idx_start = find_nearest_point(start, self.time)
        idx_end = find_nearest_point(end, self.time)

        kinds = self.textbox.get().split(", ")
        self.flight_information["segments"].append(
            dict(
                kinds=kinds,
                name="",
                irregularities=[],
                segment_id=self.flight_information["flight_id"] + "_",
                start=self.time[idx_start],
                end=self.time[idx_end],
            )
        )

        self.flight_information["segments"].sort(key=lambda x: x["start"])

        return


def flight_information(ds):
    date = datetime.datetime.strptime(ds.attrs["data_date"], "%Y%m%d").date()
    start_time = datetime.datetime.strptime(
        ds.attrs["time_coverage_start"], "%H:%M:%S UTC"
    ).time()
    end_time = datetime.datetime.strptime(
        ds.attrs["time_coverage_end"], "%H:%M:%S UTC"
    ).time()

    start_time = datetime.datetime.combine(date, start_time)
    end_time = datetime.datetime.combine(date, end_time)

    flight_number = int(ds.attrs["flight_number"])
    return dict(
        name="RF{:02d}".format(flight_number - 329),
        mission="EUREC4A",
        platform="TO",
        flight_id="TO-{:04d}".format(flight_number),
        contacts=[],
        date=date,
        flight_report="",
        takeoff=start_time,
        landing=end_time,
        events=[],
        remarks=[ds.attrs["comment"]],
        segments=[],
    )


def find_nearest_point(value, points):
    return int(np.argmin(np.abs(value - points)))


# Zeroth datetime in twinotter MASIN files
# TODO: This is not the same as the first files we had because it was updated to be
# consistent with EUREC4A. We should be able to read this from the file/dataset rather
# than assume it
t0 = datetime.datetime(1970, 1, 1)


def _convert_wacky_date_format(wacky_time):
    # The twinotter MASIN data is loaded in with a datetime coordinate but when this is
    # used on the interactive plot the value returned from the click is in days from the
    # "zeroth" datetime. Use this zeroth datetime (t0) to get the date again.
    return t0 + datetime.timedelta(days=wacky_time)


if __name__ == "__main__":
    main()



import re


def parse_docopt_arguments(function, __doc__):
    """

    Args:
        function:
        __doc__: The docstring of the function being parsed

    Returns:

    """
    # Load in the arguments
    import docopt

    arguments = docopt.docopt(__doc__)

    # Remove the help argument
    del arguments["--help"]

    # Parse the remaining arguments
    parsed_arguments = {}
    for arg in arguments.keys():
        # Don't include arguments that have not been specified at the command line to
        # keep the default arguments defined by the function (docopt doesn't allow us
        # to specify default arguments unless it is part of the options list).
        if arguments[arg] is not None:
            # Arguments specified as all upper case (<ARG>)
            if arg.upper() == arg:
                newarg = arg.lower()
            # Arguments specified in angle brackets (<arg>)
            elif re.match(r"<(.*)>", arg):
                newarg = re.match(r"<(.*)>", arg).groups()[0]
            # Arguments specificied with dashes (-arg | --arg)
            elif re.match(r"-+(.*)", arg):
                newarg = re.match(r"-+(.*)", arg).groups()[0]
            # Arguments specified by their presence (True/False arguments)
            elif arguments[arg] is True or arguments[arg] is False:
                newarg = arg
            else:
                raise KeyError("Couldn't parse argument {}".format(arg))

            parsed_arguments[newarg] = arguments[arg]

    # Call the function and return
    return function(**parsed_arguments)



"""
Check a directory for core_masin netCDF files and generate a summary .csv file
of the flights in these files

> python -m twinotter.summary /path/to/data /path/to/summary.csv
"""
import datetime
from pathlib import Path

import parse
import pandas as pd
import xarray as xr

from . import generate_file_path, MASIN_CORE_FORMAT

time_format = "{hours:02d}:{minutes:02d}:{seconds:02d} UTC"


def main():
    import argparse

    argparser = argparse.ArgumentParser()
    argparser.add_argument("flight_data_path")
    argparser.add_argument("flight_summary_path")

    args = argparser.parse_args()

    generate(
        flight_data_path=args.flight_data_path,
        flight_summary_path=args.flight_summary_path,
    )

    return


def generate(flight_data_path, flight_summary_path):
    # Get a list of netCDF files in the obs/ folder which match the expected
    # file pattern
    fn_pattern = MASIN_CORE_FORMAT.format(
        date="*", revision="*", flight_num="*", freq="*"
    )

    file_paths = list(Path(flight_data_path).rglob(fn_pattern))

    # Check if the flight summary file already exists
    if Path(flight_summary_path).exists():
        flight_summary = pd.read_csv(Path(flight_summary_path))

        for index, entry in flight_summary.iterrows():
            # Get the filename of existing entries in flight_information.csv
            file_path = generate_file_path(
                flight_number=entry["Flight Number"],
                date=entry["Date"],
                frequency=entry["Frequency"],
                revision=entry["Revision"],
                flight_data_path=flight_data_path,
            )

            # Check which files are already contained in the .csv
            if file_path in file_paths:
                print("{} already in .csv".format(file_path.name))
                file_paths.remove(file_path)

            # Other files in the .csv but not in the folder
            else:
                print("{} not available".format(file_path.name))

    else:
        flight_summary = pd.DataFrame(
            columns=["Flight Number", "Date", "Start", "End", "Revision", "Frequency"]
        )

    # Remaining files are present but not in the .csv
    # Add these new files to the .csv
    for path in file_paths:
        # Get date, flight number, revision and frequency from the filename
        flight_info = parse.parse(MASIN_CORE_FORMAT, path.name)

        # Format the date so it is more readable in the csv
        date = flight_info["date"]
        YYYY, MM, DD = int(date[0:4]), int(date[4:6]), int(date[6:8])
        date = datetime.datetime(YYYY, MM, DD)

        # Extract flight start and end from the netCDF file
        dataset = xr.open_dataset(path, decode_times=False)
        start = extract_time(dataset, "time_coverage_start")
        end = extract_time(dataset, "time_coverage_end")

        # Add flight information to .csv
        flight_summary = flight_summary.append(
            {
                "Flight Number": int(flight_info["flight_num"]),
                "Date": date.strftime("%Y-%m-%d"),
                "Start": str(start),
                "End": str(end),
                "Revision": int(flight_info["revision"]),
                "Frequency": int(flight_info["freq"]),
            },
            ignore_index=True,
        )

    flight_summary.sort_values("Flight Number", inplace=True)
    print(flight_summary)

    # Overwrite the old csv
    flight_summary.to_csv(flight_summary_path, index=False)

    return


def extract_date(dataset):
    # Get date, flight number, revision and frequency from the filename
    flight_info = parse.parse(MASIN_CORE_FORMAT, dataset.attrs["source_file"].name)

    date = flight_info["date"]
    YYYY, MM, DD = int(date[0:4]), int(date[4:6]), int(date[6:8])
    return datetime.datetime(YYYY, MM, DD)


def extract_time(dataset, name):
    time = dataset.attrs[name]

    # parse returns a dictionary when the format string has named entries.
    # use .named to get the dictionary and pass directly as keyword arguments
    # to datetime.timedelta
    return datetime.timedelta(**parse.parse(time_format, time).named)


if __name__ == "__main__":
    main()



import numpy as np
import scipy.constants

import xarray as xr
from metpy import constants
import metpy.calc


# TODO: As we add more functions to this there will be multiple paths to calculate
# a single variable. Need to decide a way of figuring out the preferred path. e.g.
# the current specific humidity calculation uses the LICOR data
# TODO: add an avoid/ignore argument so that we can avoid calculations using broken
# measurements
# TODO: Add an option to append intermediate results to the dataset
def calculate(name, ds):
    """Calculate a variable from the given dataset

    Args:
        name (str): The CF standard name
        ds (xarray.DataSet): The twin-otter MASIN dataset

    Returns:
        xarray.DataArray:

    Raises:
        ValueError: If the requested variable (or a variable required to calculate it)
            is not available in the dataset
    """
    # If the variable is in the dataset, just return it
    if name in ds:
        return ds[name]

    # If the variable name is in the translation table return the variable but renamed
    elif name in translation_table:
        result = ds[translation_table[name]].copy()
        _rename_xarray(result, name)
        return result

    # Otherwise calculate the requested variable using variables in the dataset
    elif name in available:
        arguments = []
        for argument_name in available[name]["arguments"]:
            arguments.append(calculate(argument_name, ds))

        result = available[name]["function"](*arguments)

        _rename_xarray(result, name)
        return result

    else:
        raise ValueError("Can not calculate {} from dataset".format(name))


def specific_humidity(dataset):

    x_h20 = dataset.H2O_LICOR
    q = (
        constants.water_molecular_weight
        * x_h20
        / (
            constants.water_molecular_weight * x_h20
            + constants.dry_air_molecular_weight * (1 - x_h20)
        )
    )

    return q


def along_track_wind(ds):
    angle = np.arctan2(ds.U_OXTS, ds.V_OXTS)
    magnitude = np.sqrt(ds.U_OXTS ** 2 + ds.V_OXTS ** 2)

    flight_angle = np.deg2rad(ds.HDG_OXTS)

    return magnitude * np.cos(angle - flight_angle)


def combine_temperatures(nondeiced_temperature, deiced_temperature):
    # Use the non-deiced temperature by default
    temperature = nondeiced_temperature.copy()

    # Replace with the deiced temperature where the non-deiced temperature is below
    # freezing
    frozen = np.where(temperature < scipy.constants.zero_Celsius)
    temperature[frozen] = deiced_temperature[frozen]

    return temperature


def _rename_xarray(array, name):
    array.rename(name)
    array.attrs["standard_name"] = name

    if "long_name" in array.attrs:
        del array.attrs["long_name"]


def _pint_to_xarray(quantity, ds, name):
    """The metpy functions return pint quantities but we want a DataArray consistent
    with the input DataSet

    Args:
        quantity (pint.Quantity):
        ds (xarray.DataSet):
        name (str):

    Returns:
        xarray.DataArray:
    """
    return xr.DataArray(
        data=quantity.magnitude,
        coords=ds.coords,
        dims=ds.dims,
        name=name,
        attrs=dict(
            long_name=name,
            units=quantity.units.format_babel(),
        ),
        indexes=ds.indexes,
    )


# A dictionary mapping variables that can be calculated to the functions to calculate
# them and arguments required as input to those functions
available = dict(
    air_temperature=dict(
        function=combine_temperatures,
        arguments=["TAT_ND_R", "TAT_DI_R"],
    ),
    air_potential_temperature=dict(
        function=metpy.calc.potential_temperature,
        arguments=["air_pressure", "air_temperature"],
    ),
    equivalent_potential_temperature=dict(
        function=metpy.calc.equivalent_potential_temperature,
        arguments=["air_pressure", "air_temperature", "dew_point_temperature"],
    ),
    humidity_mixing_ratio=dict(
        function=metpy.calc.mixing_ratio_from_relative_humidity,
        arguments=["air_pressure", "air_temperature", "relative_humidity"],
    ),
    relative_humidity=dict(
        function=metpy.calc.relative_humidity_from_dewpoint,
        arguments=["air_temperature", "dew_point_temperature"],
    ),
    virtual_potential_temperature=dict(
        function=metpy.calc.virtual_temperature,
        arguments=["air_potential_temperature", "humidity_mixing_ratio"],
    ),
)

# Mapping from twin-otter variable names to CF standard names. This was generated from
# an existing MASIN file. It doesn't work to just look at the standard names in the
# DataSet because some standard names are duplicated and have been removed from this
# mapping. e.g. air_temperature refers to temperature from both the deiced and
# non-deiced measurements so needs to be calculated separately
translation_table = dict(
    dew_point_temperature="TDEW_BUCK",
    brightness_temperature="BTHEIM_U",
    air_pressure="PS_AIR",
    # Q_AIR has no standard name
    downwelling_shortwave_flux_in_air="SW_DN_C",
    upwelling_shortwave_flux_in_air="SW_UP_C",
    downwelling_longwave_flux_in_air="LW_DN_C",
    upwelling_longwave_flux_in_air="LW_UP_C",
    # relative_humidity="HUM_ROSE",
    mole_fraction_of_water_vapor_in_air="H2O_LICOR",
    mole_fraction_of_carbon_dioxide_in_air="CO2_LICOR",
    flow_rate="L_FLOW_S",
    # CPC_CONC has no standard name
    latitude="LAT_OXTS",
    longitude="LON_OXTS",
    altitude="ALT_OXTS",
    platform_speed_wrt_ground_northward="VELN_OXTS",
    platform_speed_wrt_ground_eastward="VELE_OXTS",
    platform_speed_wrt_ground_upward="VELZ_OXTS",
    platform_roll_angle="ROLL_OXTS",
    platform_pitch_angle="PTCH_OXTS",
    platform_yaw_angle="HDG_OXTS",
    northward_wind="V_OXTS",
    eastward_wind="U_OXTS",
    upward_air_velocity="W_OXTS",
)

# Duplicate CF standard names
# TODO: write functions to combine these observations

# platform_speed_wrt_air="TAS_DI_AIR",
# platform_speed_wrt_air="TAS_ND_AIR",
# platform_speed_wrt_air="TAS",

# height="HGT_RADR1",
# height="HGT_RADR2",



#!/usr/bin/env python

import setuptools

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [
    "xarray",
    "numpy",
    "matplotlib",
    "pandas",
    "metpy>=1.0",
    "cartopy",
    "parse",
    "tqdm",
    "cftime",
    "worldview_dl",
    "docopt",
    "netcdf4",
    "pyyaml",
]

setup_requirements = []

test_requirements = ["pytest"]

setuptools.setup(
    author="eurec4a",
    author_email="",
    classifiers=[
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
    ],
    description="Software for working with data from the BAS twin otter",
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    include_package_data=True,
    keywords="",
    name="twinotter",
    packages=setuptools.find_packages(
        include=[
            "twinotter",
            "twinotter.plots",
            "twinotter.data",
            "twinotter.util",
            "twinotter.external",
            "twinotter.external.eurec4a",
            "twinotter.external.goes",
        ]
    ),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="",
    version="0.3.0",
)



"""Download all GOES satellite images during the given flight.

Data is downloaded to the current working directory

Usage:
    download_matching.py  <flight_data_path>
    download_matching.py  (-h | --help)

Arguments:
    <flight_data_path>  Input flight data

Options:
    -h --help        Show help

"""
import pytz

import worldview_dl

from .. import goes
from ... import load_flight, summary, util
from ...util import scripting


def main():
    scripting.parse_docopt_arguments(get_images, __doc__)
    return


def get_images(
    flight_data_path,
    layer=goes.default_layer,
    image_resolution=goes.default_spatial_resolution,
    bbox=goes.default_bbox,
):
    """Download all GOES images during the time of the specified flight

    Args:
        flight_data_path: The path to the MASIN flight data file or directory
        layer: The GOES image layer to use. Also used as the prefix of the filename
        image_resolution (float):
        bbox: The spatial area to use in the format [S W N E] with units degrees
    """
    dataset = load_flight(flight_data_path)

    date = summary.extract_date(dataset)
    date = date.astimezone(pytz.utc)

    start = date + summary.extract_time(dataset, "time_coverage_start")
    end = date + summary.extract_time(dataset, "time_coverage_end")

    start = util.round_datetime(start, goes.time_resolution)
    end = util.round_datetime(end, goes.time_resolution) + goes.time_resolution

    time = start
    while time <= end:
        # Create the output filename
        filename = goes.filename_at_time(time, layer=layer)
        print(filename)

        worldview_dl.download_image(
            fn=filename,
            time=time,
            bbox=bbox,
            layers=[layer, "Reference_Labels"],
            image_format="tiff",
            resolution=image_resolution,
        )
        time += goes.time_resolution

    return


if __name__ == "__main__":
    main()



from pathlib import Path
import re

import yaml
import xarray as xr


# netCDF naming: core_masin_YYYYMMDD_rNNN_flightNNN_Nhz.nc
MASIN_CORE_FORMAT = "core_masin_{date}_r{revision}_flight{flight_num}_{freq}hz.nc"
MASIN_CORE_RE = "core_masin_(?P<date>\d{8})_r(?P<revision>\d{3})_flight(?P<flight_num>\d{3})_(?P<freq>\d+)hz\.nc"

# A nice way of formatting the flight time
time_of_day_format = "{hours:02d}:{minutes:02d}:{seconds:02d}"


def _monkey_patch_xr_load():
    # by the CF-convections units should always be a string
    # http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.html#units
    # fix here as it breaks cf-convention loading in xarray otherwise

    def _fix_var(var):
        if var.attrs["units"] == 1:
            var.attrs["units"] = "1"

    _decode_variable_old = xr.conventions.decode_cf_variable

    def _decode_cf_variable(name, var, *args, **kwargs):
        _fix_var(var)
        return _decode_variable_old(name=name, var=var, *args, **kwargs)

    xr.conventions.decode_cf_variable = _decode_cf_variable

    _decode_variables_old = xr.conventions.decode_cf_variables

    def _decode_cf_variables(variables, *args, **kwargs):
        for var in variables:
            _fix_var(variables[var])
        return _decode_variables_old(variables, *args, **kwargs)

    xr.conventions.decode_cf_variables = _decode_cf_variables


def load_flight(flight_data_path, frequency=1, revision="most_recent", debug=False):
    # If a path to a netCDF file is specified just load it
    if Path(flight_data_path).is_file():
        meta = re.match(MASIN_CORE_RE, Path(flight_data_path).name).groupdict()
        return open_masin_dataset(flight_data_path, meta, debug=debug)

    # Otherwise a directory is supplied so look for files that match within
    # the given directory

    # If you want to use the most recent revision, get the filenames of all the
    # revisions and choose the newest one
    if revision == "most_recent":
        revision = "*"
    else:
        revision = "{:03d}".format(revision)

    fn_pattern = MASIN_CORE_FORMAT.format(
        date="*", revision=revision, flight_num="*", freq=frequency
    )
    files = list((Path(flight_data_path) / "MASIN").glob(fn_pattern))

    meta = {}
    for file in files:
        meta[file] = re.match(MASIN_CORE_RE, file.name).groupdict()

    if len(files) == 0:
        raise FileNotFoundError(
            "Couldn't find MASIN data in `{}/MASIN`, "
            "please place data there.".format(flight_data_path)
        )

    if len(files) > 1:
        if revision == "*":
            filename = sorted(files, key=lambda v: meta[v]["revision"], reverse=True)[0]
        else:
            raise FileExistsError(
                "More than one MASIN file was found: `{}`".format(", ".join(files))
            )
    else:
        filename = files[0]

    ds = open_masin_dataset(filename, meta[filename], debug=debug)

    return ds


def open_masin_dataset(filename, meta, debug=False):
    _monkey_patch_xr_load()
    ds = xr.open_dataset(filename, decode_cf=True)
    if debug:
        print("Loaded {}".format(filename))

    # drop points where lat/lon aren't given (which means the flag is 0
    # "quality_good"
    ds = ds.where(ds.LON_OXTS_FLAG == 0, drop=True)
    # drop nans too...
    ds = ds.where(~ds.LON_OXTS.isnull(), drop=True)

    # plot as function of time
    ds = ds.swap_dims(dict(data_point="Time"))

    ds.attrs["source_file"] = filename
    ds.attrs["flight_number"] = meta["flight_num"]

    return ds


def load_segments(filename):
    """Read a segments yaml file created with twinotter.plots.interactive_flight_track

    Args:
        filename (str):

    Returns:
        dict:
    """
    with open(filename, "r") as data:
        segments = yaml.load(data, yaml.CLoader)

    return segments


def _matching_segments(segments, segment_type):
    return [seg for seg in segments["segments"] if segment_type in seg["kinds"]]


def count_segments(segments, segment_type):
    """Return the number of flight segments of the requested segment_type

    Args:
        segments (dict): Flight segments description from load_segments
        segment_type (str): The label of a segment type

    Returns:
        int:

    """
    return len(_matching_segments(segments, segment_type))


def extract_segments(ds, segments, segment_type, segment_idx=None):
    """Extract a subset of the given dataset with the segments requested

    Args:
        ds (xarray.DataSet): Flight dataset
        segments (dict): Flight segments description from load_segments
        segment_type (str): The label of a segment type
        segment_idx (int): The index of the segment within the flight (starts at zero)
            If the default of None is given then the returned dataset will contain all
            matching segments concatenated.

    Returns:
        xarray.DataSet:

    """
    # All segments of the requested type
    matching_segments = _matching_segments(segments, segment_type)

    # If a single index is requested return that index of legs with the requested type
    if segment_idx is not None:
        segment = matching_segments[segment_idx]
        return ds.sel(Time=slice(segment["start"], segment["end"]))

    # Otherwise merge all legs with the requested type
    else:
        ds_matching = []
        for segment in matching_segments:
            ds_matching.append(ds.sel(Time=slice(segment["start"], segment["end"])))

        return xr.concat(ds_matching, dim="Time")


def generate_file_path(
    flight_number, date, frequency=1, revision=1, flight_data_path=None
):
    # Make the filename
    filename = MASIN_CORE_FORMAT.format(
        flight_num=flight_number,
        date=date.replace("-", ""),
        freq=frequency,
        revision="{:03d}".format(revision),
    )

    # Find all matching files in the obs directory
    file_path = list(Path(flight_data_path).rglob(filename))

    # Should only be one of these files
    if len(file_path) == 1:
        return file_path[0]
    elif len(file_path) > 1:
        raise FileExistsError("Multiple files found matching {}".format(filename))
    else:
        raise FileNotFoundError("No files found matching {}".format(filename))
